function ChainGenerator

W = 5*10^4;
numTasks = [10, 20, 30, 40, 50, 60, 70, 80, 90, 100];
regular_distributions = cell(3, 1);
regular_distributions{1} = 'Uniform';
regular_distributions{2} = 'Decrease';
regular_distributions{3} = 'Increase';

random_distributions = cell(2, 1);
random_distributions{1} = 'UniRandom';
random_distributions{2} = 'ExpRandom';
times = 10;

CR = [0.1, 1, 10];
V = [0.01, 0.1];

[num_reg_dist, x] = size(regular_distributions);
[num_rand_dist, x] = size(random_distributions);

for i = 1:length(numTasks)
   for j = 1:length(CR)
       for k = 1:length(V)
            for d = 1:num_reg_dist 
               n = numTasks(i);
               cr = CR(j);
               v = V(k);
               dist = regular_distributions(d, :);
                             
               filename = strcat('n=', num2str(n), '_', char(dist), '_CR=', num2str(cr), '_V=', num2str(v), '.chain');
               
               fid = fopen(filename, 'w');            
               if fid ~= -1
                  fprintf(fid,'%d\n', n);
                  for x = 1:n
                      % work
                      if strcmp(dist, 'Uniform')
                          w = ceil(W/n);
                      elseif strcmp(dist, 'Decrease')
                          w = (n-x+1)^2 * W / (n^3/3 + n^2/2+n/6);
                      elseif strcmp(dist, 'Increase')
                          w = x^2 * W / (n^3/3 + n^2/2+n/6);
                      else
                          disp('Unknown workload distribution')
                      end
                      fprintf(fid,'%d\t', w);    
                      % verification
                      fprintf(fid,'%d\t', w*v);    
                      % checkpoint
                      fprintf(fid,'%d\t', w*cr);   
                      % recovery
                      fprintf(fid,'%d\n', w*cr);   
                  end
                  
                  fclose(fid); 
               end
           end
       end
   end
end


for i = 1:length(numTasks)
   for j = 1:length(CR)
       for k = 1:length(V)
           for d = 1:num_rand_dist 
              for t = 1:times
                   n = numTasks(i);
                   cr = CR(j);
                   v = V(k);
                   dist = random_distributions(d, :);
                   
                   mu = W/n;
                   uni_random_nums = rand(1, n);
                   exp_random_nums = exprnd(mu, 1, n);
                   uni_random_sum = sum(uni_random_nums);
                   exp_random_sum = sum(exp_random_nums);

                   filename = strcat('n=', num2str(n), '_', char(dist), '_CR=', num2str(cr), '_V=', num2str(v), '_t=', num2str(t), '.chain');
                   fid = fopen(filename, 'w');            
                   if fid ~= -1
                      fprintf(fid,'%d\n', n);
                      for x = 1:n
                          % work
                          if strcmp(dist, 'UniRandom')
                              w = uni_random_nums(x)*W/uni_random_sum;
                          elseif strcmp(dist, 'ExpRandom')
                              w = exp_random_nums(x)*W/exp_random_sum;
                          else
                              disp('Unknown workload distribution')
                          end
                          fprintf(fid,'%d\t', w);    
                          % verification
                          fprintf(fid,'%d\t', w*v);    
                          % checkpoint
                          fprintf(fid,'%d\t', w*cr);   
                          % recovery
                          fprintf(fid,'%d\n', w*cr);   
                      end

                      fclose(fid); 
                   end
              end
           end
       end
   end
end

end