#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cmath>
#include <ctime>

#include <sys/time.h>

#include "main.hpp"

int **MEMVCOUNT1;
int **MEMVCOUNT2;

double **MEMVC;
double ****MEMV;
double ****MEMVfirst;

double **MEMEVC;
double ****MEMEV;

double *w, *v, *c, *r;

int flag = 0;

double sio = 0;
double pidle = 0;
double alpha = 0;
double beta = 0;

inline double eio_f()
{
	return alpha;
}

inline double ecpu_f(double speed)
{
	return beta * speed * speed * speed;
}

void reset(int n, int ks)
{
	for(int a=0; a<=n; ++a)
	{
		for(int b=0; b<=n; ++b)
		{
			MEMVC[a][b] = 0;
			MEMEVC[a][b] = 0;
			MEMVCOUNT1[a][b] = 0;
			MEMVCOUNT2[a][b] = 0;
			for(int c=0; c<=n; ++c)
			{
				for(int d=0; d<ks; ++d)
				{
					MEMV[a][b][c][d] = 0;
					MEMEV[a][b][c][d] = 0;
					MEMVfirst[a][b][c][d] = 0;
				}
			}
		}
	}
}

double W(int i, int j)
{
	int sum = 0;
	for(int t=i; t<=j; t++)
		sum += w[t];
	return sum;
}

double TV(int i, int j, int lc, int kv, speed sp, double &e)
{
	double s = sp.s;
	double ls = sp.ls;
	double lf = sp.lf;

	double res = 0;

	double TIJ = W(i, j) / s;

	double errS = exp(ls * TIJ);
	double errF = exp(lf * TIJ);
	double errFS = exp((lf+ls)*TIJ);

	double tmp = MEMV[lc][i-1][kv][sp.k];

	double TVj = v[j]/s;
	double TRlc = r[lc]/sio;

	double eio = eio_f();
	double ecpu = ecpu_f(s);

	double EIJ = TIJ * ecpu + TIJ * pidle;
	double EVj = TVj * ecpu + TVj * pidle;
	double ERlc = TRlc * eio + TRlc * pidle;

	double etmp = MEMEV[lc][i-1][kv][sp.k];

	if(ls == 0)
	{
		res = (errF - 1) * (1/lf + TRlc + tmp) + TVj;
		e = (errF - 1) * ((1/lf)*ecpu + ERlc + etmp) + EVj;
	}
	else if(lf == 0)
	{
		res = errS * (TIJ + TVj) + (errS - 1) * (TRlc + tmp);
		e = errS * (EIJ + EVj) + (errS - 1) * (ERlc + etmp);
	}
	else
	{
		res = errS * (((errF - 1)/lf) + TVj) + (errFS-1) * (TRlc + tmp);
		e = errS * (((errF - 1)/lf)*(ecpu + pidle) + EVj) + (errFS-1) * (ERlc + etmp);
	}

	return res;
}

double TimeVfirst(int i, int j, int kv, speed sp)
{
	if(MEMVfirst[i][j][kv][sp.k] > 0)
		return MEMVfirst[i][j][kv][sp.k];

	double s1 = sp.s;
	double ls = sp.ls;
	double lf = sp.lf;

	if(kv == 0)
	{
		double T_i_j = W(i+1, j) / s1;
		double errF_i_j = 1-exp(-lf * T_i_j);

		double ifNoFailStop = (1-errF_i_j) * (T_i_j);
		double ifFailStop = errF_i_j * (1/lf - T_i_j / (exp(lf * T_i_j)-1));

		MEMVfirst[i][j][kv][sp.k] = ifNoFailStop + ifFailStop;
		return MEMVfirst[i][j][kv][sp.k];
	}

	double min = INF;
	for(int lv=i+kv; lv<j; ++lv)
	{
		double TV = v[j] / s1;

		double T_i_lv = W(i+1, lv) / s1;
		double T_lv_j = W(lv+1, j) / s1;

		double errF_i_lv = 1-exp(-lf * T_i_lv);
		double errS_i_lv = 1-exp(-ls * T_i_lv);

		double errF_lv_j = 1-exp(-lf * T_lv_j);

		double ifNoFailStop = (1-errF_lv_j) * (T_lv_j + TV);
		double ifFailStop = errF_lv_j * (1/lf - T_lv_j / (exp(lf * T_lv_j)-1));

		double res = TimeVfirst(i, lv, kv-1, sp);
		res += (1-errF_i_lv) * (1-errS_i_lv) * (ifNoFailStop + ifFailStop);

		if(res < min)
			min = res;
	}

	MEMVfirst[i][j][kv][sp.k] = min;
	return min;
}

double TimeV(int i, int j, int kv, speed sp, double &e)
{
	if(MEMV[i][j][kv][sp.k] > 0)
	{
		e = MEMEV[i][j][kv][sp.k];
		return MEMV[i][j][kv][sp.k];
	}

	if(kv == 0)
	{
		MEMV[i][j][kv][sp.k] = TV(i+1, j, i, 0, sp, e);
		MEMEV[i][j][kv][sp.k] = e;
		return MEMV[i][j][kv][sp.k];
	}

	e = INF;
	double min = INF;
	for(int lv=i+kv; lv<j; ++lv)
	{
		double eprec = 0, esucc = 0;
		double res = TimeV(i, lv, kv-1, sp, eprec) + TV(lv+1, j, i, kv-1, sp, esucc);
		
		if(eprec + esucc < e)
		{
			min = res;
			e = eprec + esucc;
		}
	}

	MEMEV[i][j][kv][sp.k] = e;
	MEMV[i][j][kv][sp.k] = min;
	return min;
}

double TVCfirst(int i, int j, speed sp, int &nv)
{
	double min = INF;
	for(int kv=0; kv<j; ++kv)
	{
		double res = TimeVfirst(i, j, kv, sp);
		if(res < min)
		{
			min = res;
			nv = kv;
		}
	}

	return min;
}

double TVCsec(int i, int j, speed sp, int &nv, double &e)
{
	e = INF;
	double min = INF;
	for(int kv=0; kv<j; ++kv)
	{
		double etmp = 0;
		double res = TimeV(i, j, kv, sp, etmp);
		if(etmp < e)
		{
			min = res;
			nv = kv;
			e = etmp;
		}
	}

	return min;
}

double TVC(int i, int j, speed sp1, speed sp2, int &nv1, int &nv2, double &e)
{
	double s1 = sp1.s;
	double lf = sp1.lf;
	double ls = sp1.ls;

	double TIJ = W(i+1, j) / s1;
	double TC = c[j]/sio;
	double TR = r[i]/sio;

	double errF = 1-exp(-lf * TIJ);
	double errS = 1-exp(-ls * TIJ);

	double esec = 0;
	double res1 = TVCfirst(i, j, sp1, nv1);
	double res2 = TVCsec(i, j, sp2, nv2, esec);

	double eio = eio_f();
	double ecpu = ecpu_f(s1);

	double eprec = res1 * ecpu + res1 * pidle;

	double ER = TR * eio + TR * pidle;
	double EC = TC * eio + TC * pidle;

	double res = res1 + ((1-errF) * errS + errF) * (TR + res2) + TC;

	e = eprec + ((1-errF) * errS + errF) * (ER + esec) + EC;

	return res;
}

double TVCBestSpeeds(int i, int j, int ks, speed *spList, int &nv1, int &nv2, double &e)
{
	e = INF;
	double min = INF;
	int nvmin1 = 0, nvmin2 = 0;
	for(int s1=0; s1<ks; ++s1)
	{
		for(int s2=0; s2<ks; ++s2)
		{
			double etmp = 0;
			double res = TVC(i, j, spList[s1], spList[s2], nvmin1, nvmin2, etmp);
			if(etmp < e)
			{
				e = etmp;
				min = res;
				nv1 = nvmin1;
				nv2 = nvmin2;
			}
		}
	}

	return min;
}

double TimeVC(int j, int k, int ks, speed *spList, int &nv1, int &nv2, double &e)
{
	if(MEMVC[j][k] > 0)
	{
		nv1 = MEMVCOUNT1[j][k];
		nv2 = MEMVCOUNT2[j][k];
		e = MEMEVC[j][k];
		return MEMVC[j][k];
	}

	int nvmin1 = 0, nvmin2 = 0;
	if(k == 0)
	{	
		MEMVC[j][k] = TVCBestSpeeds(0, j, ks, spList, nvmin1, nvmin2, e);
		MEMEVC[j][k] = e;
		MEMVCOUNT1[j][k] = nv1 = nvmin1;
		MEMVCOUNT2[j][k] = nv2 = nvmin2;

		return MEMVC[j][k];
	}

	int nvprec1 = 0, nvprec2 = 0;
	int nvsucc1 = 0, nvsucc2 = 0;

	e = INF;
	double min = INF;
	for(int i=k; i<j; ++i)
	{
		double eprec = 0, esucc = 0;
		double prec = TimeVC(i, k-1, ks, spList, nvprec1, nvprec2, eprec);
		double succ = TVCBestSpeeds(i, j, ks, spList, nvsucc1, nvsucc2, esucc);

		if(eprec + esucc < e)
	 	{
	 		min = prec + succ;
	 		e = eprec + esucc;
			nvmin1 = nvprec1 + nvsucc1;
			nvmin2 = nvprec2 + nvsucc2;
	 	}
	}

	MEMVC[j][k] = min;
	MEMEVC[j][k] = e;
	MEMVCOUNT1[j][k] = nv1 = nvmin1;
	MEMVCOUNT2[j][k] = nv2 = nvmin2;

	return min;
}

double computeTime(int n, int ks, speed *spList, int &nckpt, int &nv1, int &nv2, double &e)
{
	reset(n, ks);

	e = INF;
	double min = INF;
	int nvmin1 = 0, nvmin2 = 0;
	for(int k=0; k<n; ++k)
	{
		double etmp = 0;
		double res = TimeVC(n, k, ks, spList, nvmin1, nvmin2, etmp);

		printf("%d %d %d %d = %lf, %lf\n", n, k, nvmin1 + k, nvmin2 + k, res, etmp);
		if(etmp < e)
		{
			min = res;
			e = etmp;
			nckpt = k;
			nv1 = nvmin1 + k;
			nv2 = nvmin2 + k;
		}
	}

	return min;
}

bool launch(const char *filename, int n, int ks, double *speed, double *lambdaF, double *lambdaS)
{
	FILE *out = fopen(filename, "w");
	if(out)
	{
		struct speed *spList = new struct speed[ks];

		for(int i=0; i<ks; ++i)
		{
			spList[i].k = i;
			spList[i].s = speed[i];
			spList[i].lf = lambdaF[i];
			spList[i].ls = lambdaS[i];
		}

		double etmp = 0;
		int nckpt = 0, nv1 = 0, nv2 = 0;
		double minTime = computeTime(n, ks, spList, nckpt, nv1, nv2, etmp);

		fprintf(out, "%lf %d %d %d %lf\n", minTime, nckpt, nv1, nv2, etmp);

		delete[] spList;

		return true;
	}
	else
		return false;
}

bool read_params(const char *filename, double *pidle, double *alpha, double *sio, double *beta, int *ks, double **speed, double **lambdaF, double **lambdaS)
{
	FILE *file = fopen(filename, "r");
	if(file)
	{
		char target[16];
		fscanf(file, "%s\n%lf\n%lf\n%lf\n%lf\n%d\n", target, pidle, alpha, sio, beta, ks);

		if(strcmp(target, "C") == 0)
			flag = 0;
		else if(strcmp(target, "VC") == 0)
			flag = 1;

		*speed = new double[*ks];
		*lambdaF = new double[*ks];
		*lambdaS = new double[*ks];

		for(int i=0; i<*ks; ++i)
			fscanf(file, "%lf", &(*speed)[i]);

		for(int i=0; i<*ks; ++i)
			fscanf(file, "%lf", &(*lambdaF)[i]);

		for(int i=0; i<*ks; ++i)
			fscanf(file, "%lf", &(*lambdaS)[i]);

		fclose(file);

		return true;
	}
	else
		return false;
}

bool read_chain(const char *filename, int *n, double **w, double **v, double **c, double **r)
{
	FILE *file = fopen(filename, "r");
	if(file)
	{
		fscanf(file, "%d\n", n);

		if(*n > 100)
		{
			printf("N too big: %d\n", *n);
			fclose(file);
			return false;
		}

		*w = new double[*n+1];
		*v = new double[*n+1];
		*c = new double[*n+1];
		*r = new double[*n+1];

		(*w)[0] = (*v)[0] = (*c)[0] = (*r)[0] = 0;

		for(int i=1; i<=*n; ++i)
			fscanf(file, "%lf %lf %lf %lf\n", &(*w)[i], &(*v)[i], &(*c)[i], &(*r)[i]);

		fclose(file);

		return true;
	}
	else
		return false;
}

int main(int argc, char **argv)
{
	if(argc < 4)
	{
		printf("usage: %s params chain output\n", argv[0]);
		return 0;
	}

	struct timeval time; 
    gettimeofday(&time,NULL);
    srand((time.tv_sec * 1000) + (time.tv_usec / 1000));

    int n, ks;
    double *speed, *lambdaF, *lambdaS;

    if(read_params(argv[1], &pidle, &alpha, &sio, &beta, &ks, &speed, &lambdaF, &lambdaS) == false)
    {
    	printf("Unable to open %s\n", argv[1]);
    	return 1;
    }

    if(read_chain(argv[2], &n, &w, &v, &c, &r) == false)
	{
		printf("Unable to open %s\n", argv[2]);
    	return 1;
	}

	printf("Target = %s\n", flag == 0? "C" : "VC");
    printf("N = %d\n", n);
    printf("Pio = %lf\n", alpha);
    printf("Sio = %lf\n", sio);
    printf("Pidle = %lf\n", pidle);
    printf("Beta = %lf\n", beta);

    printf("speeds: ");
    for(int i=0; i<ks; ++i)
		printf("%lf ", speed[i]);
	printf("\n");

    printf("lambdas F: ");
    for(int i=0; i<ks; ++i)
		printf("%lf ", lambdaF[i]);
	printf("\n");

	printf("lambdas S: ");
    for(int i=0; i<ks; ++i)
		printf("%lf ", lambdaS[i]);
	printf("\n");

	for(int i=1; i<=n; ++i)
		printf("%lf %lf %lf %lf\n", w[i], v[i], c[i], r[i]);

	MEMVCOUNT1 = new int*[n+1];
	MEMVCOUNT2 = new int*[n+1];

	MEMVC = new double*[n+1];
	MEMEVC = new double*[n+1];
	MEMV = new double***[n+1];
	MEMEV = new double***[n+1];
	MEMVfirst = new double***[n+1];
	for(int a=0; a<=n; ++a)
	{
		MEMVCOUNT1[a] = new int[n+1];
		MEMVCOUNT2[a] = new int[n+1];

		MEMVC[a] = new double[n+1];
		MEMEVC[a] = new double[n+1];

		MEMV[a] = new double**[n+1];
		MEMEV[a] = new double**[n+1];
		MEMVfirst[a] = new double**[n+1];

		for(int b=0; b<=n; ++b)
		{
			MEMV[a][b] = new double*[n+1];
			MEMEV[a][b] = new double*[n+1];
			MEMVfirst[a][b] = new double*[n+1];

			for(int c=0; c<=n; ++c)
			{
				MEMV[a][b][c] = new double[ks];
				MEMEV[a][b][c] = new double[ks];
				MEMVfirst[a][b][c] = new double[ks];
			}
		}
	}

	if(launch(argv[3], n, ks, speed, lambdaF, lambdaS) == false)
	{
		printf("Unable to open %s\n", argv[3]);
		return 1;
	}

    for(int a=0; a<=n; ++a)
	{
		for(int b=0; b<=n; ++b)
		{
			for(int c=0; c<=n; ++c)
			{
				delete[] MEMV[a][b][c];
				delete[] MEMEV[a][b][c];
				delete[] MEMVfirst[a][b][c];
			}
			delete[] MEMV[a][b];
			delete[] MEMEV[a][b];
			delete[] MEMVfirst[a][b];
		}

		delete[] MEMVC[a];
		delete[] MEMEVC[a];

		delete[] MEMV[a];
		delete[] MEMEV[a];
		delete[] MEMVfirst[a];

		delete[] MEMVCOUNT1[a];
		delete[] MEMVCOUNT2[a];
	}

	delete[] MEMVC;
	delete[] MEMEVC;

	delete[] MEMV;
	delete[] MEMEV;
	delete[] MEMVfirst;

	delete[] MEMVCOUNT1;
	delete[] MEMVCOUNT2;

	delete[] w;
	delete[] v;
	delete[] c;
	delete[] r;

	delete[] speed;
	delete[] lambdaS;
	delete[] lambdaF;

	return 0;
}