#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cmath>
#include <ctime>

#include <sys/time.h>

#include "main.hpp"

int **MEMVCOUNT;
double **MEMEVC;
double ***MEMV;

double **MEMVC;
double ***MEMEV;

double *w, *v, *c, *r;

int flag = 0;

double sio = 0;
double pidle = 0;
double alpha = 0;
double beta = 0;

inline double eio_f()
{
	return alpha;
}

inline double ecpu_f(double speed)
{
	return beta * speed * speed * speed;
}

void reset(int n)
{
	for(int a=0; a<=n; ++a)
	{
		for(int b=0; b<=n; ++b)
		{
			MEMVC[a][b] = 0;
			MEMEVC[a][b] = 0;
			MEMVCOUNT[a][b] = 0;
			for(int c=0; c<=n; ++c)
			{
				MEMV[a][b][c] = 0;
				MEMEV[a][b][c] = 0;
			}
		}
	}
}

double W(int i, int j)
{
	int sum = 0;
	for(int t=i; t<=j; t++)
		sum += w[t];
	return sum;
}

double TV(int i, int j, int lc, int kv, double s, double lf, double ls, double &e)
{
	double res = 0;

	double TIJ = W(i, j) / s;

	double errS = exp(ls * TIJ);
	double errF = exp(lf * TIJ);
	double errFS = exp((lf+ls)*TIJ);

	double tmp = MEMV[lc][i-1][kv];

	double TVj = v[j]/s;
	double TRlc = r[lc]/sio;

	double eio = eio_f();
	double ecpu = ecpu_f(s);

	double EIJ = TIJ * ecpu + TIJ * pidle;
	double EVj = TVj * ecpu + TVj * pidle;
	double ERlc = TRlc * eio + TRlc * pidle;

	double etmp = MEMEV[lc][i-1][kv];

	if(ls == 0)
	{
		res = (errF - 1) * (1/lf + TRlc + tmp) + TVj;
		e = (errF - 1) * ((1/lf)*ecpu + ERlc + etmp) + EVj;
	}
	else if(lf == 0)
	{
		res = errS * (TIJ + TVj) + (errS - 1) * (TRlc + tmp);
		e = errS * (EIJ + EVj) + (errS - 1) * (ERlc + etmp);
	}
	else
	{
		res = errS * (((errF - 1)/lf) + TVj) + (errFS-1) * (TRlc + tmp);
		e = errS * (((errF - 1)/lf)*(ecpu + pidle) + EVj) + (errFS-1) * (ERlc + etmp);
	}

	return res;
}

float TC(int i, int j, float s, float lf, float ls)
{
	double res = 0;

	double TIJ = W(i, j) / s;

	double errS = exp(ls * TIJ);
	double errF = exp(lf * TIJ);
	double errFS = exp((lf+ls)*TIJ);

	double TR = r[i-1] / sio;
	double TC = c[j] / sio;
	double TV = v[j]/s;

	if(ls == 0)
		res = (errF-1) * ((1/lf) + TR) + TC;
	else if(lf == 0)
		res = errS * (TIJ + TV) + (errS-1) * TR + TC;
	else
		res = errS * (((errF-1)/lf) + TV) + ((errFS-1) * TR) + TC;
	
	return res;
}

double TimeV(int i, int j, int kv, double s, double lf, double ls, double &e)
{
	if(MEMV[i][j][kv] > 0)
	{
		e = MEMEV[i][j][kv];
		return MEMV[i][j][kv];
	}

	if(kv == 0)
	{
		MEMV[i][j][kv] = TV(i+1, j, i, 0, s, lf, ls, e);
		MEMEV[i][j][kv] = e;
		return MEMV[i][j][kv];
	}

	double min = INF;
	for(int lv=i+kv; lv<j; ++lv)
	{
		double eprec = 0, esucc = 0;
		double res = TimeV(i, lv, kv-1, s, lf, ls, eprec) + TV(lv+1, j, i, kv-1, s, lf, ls, esucc);

		if(res < min)
		{
			min = res;
			e = eprec + esucc;
		}
	}

	MEMEV[i][j][kv] = e;
	MEMV[i][j][kv] = min;
	return min;
}

double TimeVC(int j, int k, double s, double lf, double ls, int &nv, double &e)
{
	if(MEMVC[j][k] > 0)
	{
		nv = MEMVCOUNT[j][k];
		e = MEMEVC[j][k];
		return MEMVC[j][k];
	}

	double eio = eio_f();

	int nvmin = 0;
	if(k == 0)
	{	
		double min = INF;
		for(int kv=0; kv<j; ++kv)
		{
			double etmp = 0;
			double TC = (c[j]/sio);

			double res = TimeV(0, j, kv, s, lf, ls, etmp) + c[j]/sio;
			
			if(res < min)
			{
				min = res;
				nvmin = kv;
				e = etmp + TC * eio + TC * pidle;
			}
		}

		MEMVC[j][k] = min;
		MEMEVC[j][k] = e;
		MEMVCOUNT[j][k] = nv = nvmin;

		return MEMVC[j][k];
	}

	int nvprec = 0;
	double min = INF;
	for(int i=k; i<j; ++i)
	{
		double eprec = 0, esucc = 0;
		double TC = (c[j]/sio);

		double prec = TimeVC(i, k-1, s, lf, ls, nvprec, eprec) + c[j]/sio;

		for(int kv=0; kv<j-i; ++kv)
		{
		 	double succ = TimeV(i, j, kv, s, lf, ls, esucc);
		 	if(prec + succ < min)
		 	{
		 		min = prec + succ;
				nvmin = nvprec + kv;
				e = eprec + esucc + TC * eio + TC * pidle;
		 	}
		}
	}

	MEMVC[j][k] = min;
	MEMEVC[j][k] = e;
	MEMVCOUNT[j][k] = nv = nvmin;

	return min;
}

float TimeC(int j, int k, float s, float lf, float ls)
{
	if(MEMVC[j][k] > 0)
		return MEMVC[j][k];

	if(k == 0)
	{
		MEMVC[j][k] = TC(1, j, s, lf, ls);
		return MEMVC[j][k];
	}

	float min = INF;
	for(int i=k; i<j; ++i)
	{
		float res = TimeC(i, k-1, s, lf, ls) + TC(i+1, j, s, lf, ls);
		if(res < min)
			min = res;
	}

	MEMVC[j][k] = min;

	return min;
}

double computeTime(int n, double s, double lf, double ls, int &nckpt, int &nv, double &e)
{
	reset(n);

	int nvrec = 0;
	double min = INF;
	for(int k=0; k<n; ++k)
	{
		double res = 0, etmp = 0;
		if(flag == 0) res = TimeC(n, k, s, lf, ls);
		else if(flag == 1) res = TimeVC(n, k, s, lf, ls, nvrec, etmp);

		printf("%d %d %d %lf %lf %lf = %lf, %lf\n", n, k, nvrec + k, s, lf, ls, res, etmp);
		if(res < min)
		{
			min = res;
			nckpt = k;
			nv = nvrec + k;
			e = etmp;
		}
	}

	return min;
}

bool launch(const char *filename, int n, int ks, double *speed, double *lambdaF, double *lambdaS)
{
	FILE *out = fopen(filename, "w");
	if(out)
	{
		for(int i=0; i<ks; ++i)
		{
			double etmp = 0;
			int nckpt = 0, nv = 0;
			double minTime = computeTime(n, speed[i], lambdaF[i], lambdaS[i], nckpt, nv, etmp);
			fprintf(out, "%lf %lf %lf %lf %d %d %lf\n", speed[i], lambdaF[i], lambdaS[i], minTime, nckpt, nv, etmp);
		}

		return true;
	}
	else
		return false;
}

bool read_params(const char *filename, double *pidle, double *alpha, double *sio, double *beta, int *ks, double **speed, double **lambdaF, double **lambdaS)
{
	FILE *file = fopen(filename, "r");
	if(file)
	{
		char target[16];
		fscanf(file, "%s\n%lf\n%lf\n%lf\n%lf\n%d\n", target, pidle, alpha, sio, beta, ks);

		if(strcmp(target, "C") == 0)
			flag = 0;
		else if(strcmp(target, "VC") == 0)
			flag = 1;

		*speed = new double[*ks];
		*lambdaF = new double[*ks];
		*lambdaS = new double[*ks];

		for(int i=0; i<*ks; ++i)
			fscanf(file, "%lf", &(*speed)[i]);

		for(int i=0; i<*ks; ++i)
			fscanf(file, "%lf", &(*lambdaF)[i]);

		for(int i=0; i<*ks; ++i)
			fscanf(file, "%lf", &(*lambdaS)[i]);

		fclose(file);

		return true;
	}
	else
		return false;
}

bool read_chain(const char *filename, int *n, double **w, double **v, double **c, double **r)
{
	FILE *file = fopen(filename, "r");
	if(file)
	{
		fscanf(file, "%d\n", n);

		if(*n > 100)
		{
			printf("N too big: %d\n", *n);
			fclose(file);
			return false;
		}

		*w = new double[*n+1];
		*v = new double[*n+1];
		*c = new double[*n+1];
		*r = new double[*n+1];

		(*w)[0] = (*v)[0] = (*c)[0] = (*r)[0] = 0;

		for(int i=1; i<=*n; ++i)
			fscanf(file, "%lf %lf %lf %lf\n", &(*w)[i], &(*v)[i], &(*c)[i], &(*r)[i]);

		fclose(file);

		return true;
	}
	else
		return false;
}

int main(int argc, char **argv)
{
	if(argc < 4)
	{
		printf("usage: %s params chain output\n", argv[0]);
		return 0;
	}

	struct timeval time; 
    gettimeofday(&time,NULL);
    srand((time.tv_sec * 1000) + (time.tv_usec / 1000));

    int n, ks;
    double *speed, *lambdaF, *lambdaS;

    if(read_params(argv[1], &pidle, &alpha, &sio, &beta, &ks, &speed, &lambdaF, &lambdaS) == false)
    {
    	printf("Unable to open %s\n", argv[1]);
    	return 1;
    }

    if(read_chain(argv[2], &n, &w, &v, &c, &r) == false)
	{
		printf("Unable to open %s\n", argv[2]);
    	return 1;
	}

	printf("Target = %s\n", flag == 0? "C" : "VC");
    printf("N = %d\n", n);
    printf("Pio = %lf\n", alpha);
    printf("Sio = %lf\n", sio);
    printf("Pidle = %lf\n", pidle);
    printf("Beta = %lf\n", beta);

    printf("speeds: ");
    for(int i=0; i<ks; ++i)
		printf("%lf ", speed[i]);
	printf("\n");

    printf("lambdas F: ");
    for(int i=0; i<ks; ++i)
		printf("%lf ", lambdaF[i]);
	printf("\n");

	printf("lambdas S: ");
    for(int i=0; i<ks; ++i)
		printf("%lf ", lambdaS[i]);
	printf("\n");

	for(int i=1; i<=n; ++i)
		printf("%lf %lf %lf %lf\n", w[i], v[i], c[i], r[i]);

	MEMVCOUNT = new int*[n+1];
	MEMVC = new double*[n+1];
	MEMEVC = new double*[n+1];
	MEMV = new double**[n+1];
	MEMEV = new double**[n+1];
	for(int a=0; a<=n; ++a)
	{
		MEMVCOUNT[a] = new int[n+1];
		MEMVC[a] = new double[n+1];
		MEMEVC[a] = new double[n+1];
		MEMV[a] = new double*[n+1];
		MEMEV[a] = new double*[n+1];
		for(int b=0; b<=n; ++b)
		{
			MEMV[a][b] = new double[n+1];
			MEMEV[a][b] = new double[n+1];
		}
	}

	if(launch(argv[3], n, ks, speed, lambdaF, lambdaS) == false)
	{
		printf("Unable to open %s\n", argv[3]);
		return 1;
	}

    for(int a=0; a<=n; ++a)
	{
		for(int b=0; b<=n; ++b)
		{
			delete[] MEMV[a][b];
			delete[] MEMEV[a][b];
		}

		delete[] MEMV[a];
		delete[] MEMEV[a];
		delete[] MEMVC[a];
		delete[] MEMEVC[a];
		delete[] MEMVCOUNT[a];
	}

	delete[] MEMV;
	delete[] MEMEV;
	delete[] MEMVC;
	delete[] MEMEVC;
	delete[] MEMVCOUNT;

	delete[] w;
	delete[] v;
	delete[] c;
	delete[] r;

	delete[] speed;
	delete[] lambdaS;
	delete[] lambdaF;

	return 0;
}